from abc import ABC, abstractmethod


class I_Auth(ABC):
    """Authentification module interface
    """

    @abstractmethod
    def authenticate(self):
        """Abstract method used for authentication

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement authenticate()')



import google.auth
from cdktf_cdktf_provider_google.provider import GoogleProvider

from thipster.engine import I_Auth


class GoogleAuth(I_Auth):
    """Authenticate to Google Cloud Platform (GCP) projects

    To use this module, you need to have a GCP account and a project created. 
    You also need to have gcloud installed : https://cloud.google.com/sdk/docs/install
    Then login using glcloud : gcloud auth application-default login
    """

    def __init__(self) -> None:
        return GoogleAuth

    def authenticate(app):
        """Generates the google provider block for the Terraform CDK

        Parameters
        ----------
        app: Construct
            CDK Construct where the provider is created

        """

        credentials, project_id = google.auth.default()

        GoogleProvider(
            app, 'default_google',
            project=project_id,
            access_token=credentials.token,
        )



import json
import os
import shutil

import pytest
from cdktf_cdktf_provider_google.provider import GoogleProvider

from thipster.engine.engine import Engine
from thipster.engine.i_auth import I_Auth
from thipster.parser.parser_factory import ParserFactory
from thipster.repository.local import LocalRepo
from thipster.terraform import Terraform
from thipster.terraform.exceptions import (
    CDKCyclicDependencies,
    CDKDependencyNotDeclared,
    CDKMissingAttributeInDependency,
)

LOCAL_REPO = 'tests/resources/e2e/models'
REMOTE_REPO = 'THipster/models'


class MockAuth(I_Auth):
    def authenticate(app):
        GoogleProvider(
            app, 'default_google',
            project='project_id',
            credentials='credentials.token',
        )


def create_dir(dirname: str, files: dict[str, str]):
    if not os.path.isdir(dirname):
        os.mkdir(dirname)

    dirname = os.path.abspath(dirname)
    for name, content in files.items():
        create_file(name, content, dirname)

    def destroy_files():
        shutil.rmtree(dirname)

    return destroy_files


def create_file(filename: str, content: str, dirname: str = 'test'):
    if not os.path.isdir(dirname):
        os.mkdir(dirname)
    dirname = os.path.abspath(dirname)

    file = open(f'{dirname}/{filename}', 'w')
    file.write(content)
    file.close()


def __test_file(file: str, local_repo: str = LOCAL_REPO, file_type: str = 'thips'):
    path_input = 'test'
    __destroy_dir = create_dir(
        path_input,
        {
            f'test_file.{file_type}':
            file,
        },
    )

    engine = Engine(
        ParserFactory(),
        LocalRepo(local_repo),
        MockAuth,
        Terraform(),
    )
    try:
        output = engine.run(path_input)
    except Exception as e:
        raise e
    finally:
        __destroy_dir()

        if os.path.exists('cdktf.out'):
            shutil.rmtree('cdktf.out')

    return output


def get_output():
    with open('thipster.tf.json') as f:
        file_contents = json.load(f)
        f.close()
    return file_contents


def get_resource(resource_data: tuple):
    output = get_output()

    resources_of_type = output.get('resource').get(resource_data[0])
    resource = None
    for r in resources_of_type.values():
        if r.get('name') == resource_data[1]:
            resource = r

    assert resource is not None

    return resource


def assert_resource_created(
    resource_type: str,
    resource_name: str,
):
    output = get_output()
    assert output.get('resource') is not None
    resources = output.get('resource')

    assert resources.get(resource_type) is not None

    names = [x.get('name') for _, x in resources.get(resource_type).items()]
    assert resource_name in names

    return (resource_type, resource_name)


def assert_number_of_resource_type_is(
    resource_type: str,
    amount: str,
):
    output = get_output()
    assert output.get('resource') is not None
    resources = output.get('resource')

    assert resources.get(resource_type) is not None
    assert len(resources.get(resource_type)) == amount


def assert_resource_parameters_are(resource_data: tuple, parameters: list[str]):
    resource = get_resource(resource_data)

    for parameter in parameters:
        assert parameter in resource.keys()


def get_resource_parameter(resource_data: tuple, parameter: str):
    resource = get_resource(resource_data)

    assert parameter in resource.keys()
    return resource.get(parameter)


def test_bucket():
    __test_file(
        file="""
bucket my-bucket:
\tregion : euw
    """,
    )

    assert_number_of_resource_type_is('google_storage_bucket', 1)
    bucket = assert_resource_created('google_storage_bucket', 'my-bucket')
    assert_resource_parameters_are(bucket, ['location'])


def test_empty_bucket():
    __test_file(
        file="""
bucket dzvhvzarbazkhr:

    """,
    )

    assert_number_of_resource_type_is('google_storage_bucket', 1)
    assert_resource_created('google_storage_bucket', 'dzvhvzarbazkhr')

    __test_file(
        file="""
bucket dzvhvzarbazkhr:

bucket ezezeaz:
    region: europe
    """,
    )

    assert_number_of_resource_type_is('google_storage_bucket', 2)
    assert_resource_created('google_storage_bucket', 'dzvhvzarbazkhr')
    assert_resource_created('google_storage_bucket', 'ezezeaz')


def test_dep_with_no_options():
    with pytest.raises(CDKMissingAttributeInDependency):
        __test_file(
            file="""
bucket_bad_dep_parent my-bucket:
\tregion : euw
        """,
        )


def test_cyclic_deps():
    with pytest.raises(CDKCyclicDependencies):
        __test_file(
            file="""
bucket_bad_dep_cyclic my-bucket:
\tregion : euw
        """,
        )


def test_lb():
    __test_file(
        file="""
network lb-net:

subnetwork lb-subnet:
\tnetwork: lb-net
\tregion: europe-west1b
\tip_range: 10.0.1.0/24

loadbalancer my-lb:
\tload_balancing_scheme: EXTERNAL
    """,
    )

    assert_number_of_resource_type_is('google_compute_network', 3)
    assert_resource_created('google_compute_network', 'lb-net')

    assert_number_of_resource_type_is('google_compute_subnetwork', 1)
    assert_resource_created('google_compute_subnetwork', 'lb-subnet')


def test_internal_object():
    __test_file(
        file="""
firewall testParent:
\tdirection: EGRESS
        """,
    )

    __test_file(
        file="""
firewall testParent:
\tdirection: EGRESS

\tallow:
\t\tprotocol: http
        """,
    )


def test_missing_explicit_dependency():
    with pytest.raises(CDKDependencyNotDeclared):
        __test_file(
            file="""
subnetwork lb-subnet:
\tnetwork: lb-net
\tregion: europe-west1b
\tip_range: 10.0.1.0/24
            """,
        )


def test_explicit_dependency():
    __test_file(
        file="""
network lb-net:

subnetwork lb-subnet:
\tnetwork: lb-net
\tregion: europe-west1b
\tip_range: 10.0.1.0/24
        """,
    )


def test_bucket_cors():
    __test_file(
        file="""
bucket corsBucket:
    cors:
        origin:
            - "http://example.com"
        method:
            - "*"
        responseHeader:
            - "*"
        maxAge: 400
        """,
    )

    assert_number_of_resource_type_is('google_storage_bucket', 1)
    assert_resource_created('google_storage_bucket', 'corsBucket')


def test_bucket_two_cors():
    __test_file(
        file="""
bucket:
  name: corsBucket
  cors:
    - origin:
      - "http://example.com"
      method:
      - "*"
      responseHeader:
      - "*"
      maxAge: 400
    - origin:
      - "http://example.com/other"
      method:
      - "*"
      responseHeader:
      - "*"
      maxAge: 600
""",
        file_type='yaml',
    )

    assert_number_of_resource_type_is('google_storage_bucket', 1)
    bucket = assert_resource_created('google_storage_bucket', 'corsBucket')
    cors_block = get_resource_parameter(bucket, 'cors')

    assert len(cors_block) == 2



from .token import TOKENTYPES as TT
from .token import Token


class DSLParserBaseException(Exception):
    def __init__(self, message, *args: object) -> None:
        super().__init__(*args)

        self.__message = message

    @property
    def message(self):
        return self.__message


class DSLParserPathNotFound(DSLParserBaseException):
    def __init__(self, file, *args: object) -> None:
        super().__init__(f'Path not found : {file}', *args)


class DSLSyntaxException(Exception):
    def __init__(self, token: Token, expected: TT | list[TT], *args: object) -> None:
        super().__init__(*args)
        self.__tok = token
        self.__exp = expected

    def __repr__(self) -> str:

        if type(self.__exp) is TT:
            return f"""{str(self.__tok.position)} :\n\tSyntax error : Expected \
{str(self.__exp.value)}, got {str(self.__tok.tokenType)}"""
        else:
            return f"""{str(self.__tok.position)} :\n\tSyntax error : Expected \
{str(' or '.join(list(map(lambda x : str(x), self.__exp))))}, got {
    str(self.__tok.tokenType)}"""

    @property
    def tok(self):
        return self.__tok


class DSLConditionException(Exception):
    def __init__(self, token: Token, *args: object) -> None:
        super().__init__(*args)
        self.__tok = token

    def __repr__(self) -> str:
        return f"""{str(self.__tok.position)} :\n\tBad condition"""


class DSLUnexpectedEOF(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    def __repr__(self) -> str:
        return 'Unexpected EOF'


class DSLParserNoEndingQuotes(DSLParserBaseException):
    def __init__(self, position, *args: object) -> None:
        super().__init__(
            f'Invalid syntax, missing ending quotes at : {position}', *args,
        )



import thipster.engine.parsed_file as pf
import thipster.parser.dsl_parser.ast as ast

from .exceptions import DSLParserBaseException
from .token import TOKENTYPES as TT
from .token_parser import DSLSyntaxException


class DSLParserVariableAlreadyUsed(DSLParserBaseException):
    def __init__(self, var: str, *args: object) -> None:
        super().__init__(f'Variable already used : {var}', *args)


class DSLParserVariableNotDeclared(DSLParserBaseException):
    def __init__(self, var: str, *args: object) -> None:
        super().__init__(f'Variable already used : {var}', *args)


class Interpreter():
    """Interpreter class for the DSL Parser

    Implements a visitor design pattern on the AST nodes
    """

    def __init__(self) -> None:
        super().__init__()

        self.__variables = dict[str, int]()

    def run(self, tree: ast.FileNode) -> pf.ParsedFile:
        """Runs the interpreter on an Abstract Syntax Tree

        Parameters
        ----------
        tree: FileNode
            The root node of the abstract syntax tree

        Returns
        -------
        ParsedFile
            The parsed file stucture assiociated to the given AST
        """
        return tree.accept(self)

    def visitCompExpr(self, element: ast.CompExprNode) -> bool:
        """Visitor for a StringNode

        Parameters
        ----------
        element: StringNode
            The visited node

        Returns
        -------
        bool
            The boolean value of the node
        """
        match element.operation.tokenType:
            case TT.NOT:
                return pf.ParsedLiteral(not element.leftValue.accept(self).value)

            case TT.OR:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value or
                    element.rightValue.accept(self).value,
                )

            case TT.AND:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value and
                    element.rightValue.accept(self).value,
                )

            case TT.EE:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value ==
                    element.rightValue.accept(self).value,
                )

            case TT.NE:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value !=
                    element.rightValue.accept(self).value,
                )

            case TT.LT:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value <
                    element.rightValue.accept(self).value,
                )

            case TT.LTE:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value <=
                    element.rightValue.accept(self).value,
                )

            case TT.GT:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value >
                    element.rightValue.accept(self).value,
                )

            case TT.GTE:
                return pf.ParsedLiteral(
                    element.leftValue.accept(self).value >=
                    element.rightValue.accept(self).value,
                )

    def visitArithExpr(self, element: ast.ArithExprNode) -> int | float:
        """Visitor for a StringNode

        Parameters
        ----------
        element: StringNode
            The visited node

        Returns
        -------
        int | float
            The arithmetic value of the node
        """
        total = element.terms[0].accept(self)
        if isinstance(total, pf.ParsedLiteral):
            total = total.value

        for i in range(len(element.operation)):
            match element.operation[i]:
                case TT.PLUS:
                    add = element.terms[i+1].accept(self)
                    if isinstance(add, pf.ParsedLiteral):
                        add = add.value
                    total += add

                case TT.MINUS:
                    rm = element.terms[i+1].accept(self)
                    if isinstance(rm, pf.ParsedLiteral):
                        rm = rm.value
                    total -= rm

        return pf.ParsedLiteral(total)

    def visitTerm(self, element: ast.TermNode) -> int | float:
        """Visitor for a StringNode

        Parameters
        ----------
        element: StringNode
            The visited node

        Returns
        -------
        int | float
            The arithmetic value of the node
        """
        total = element.factors[0].accept(self)
        if isinstance(total, pf.ParsedLiteral):
            total = total.value

        for i in range(len(element.operation)):
            match element.operation[i]:
                case TT.MUL:
                    mul = element.factors[i+1].accept(self)
                    if isinstance(mul, pf.ParsedLiteral):
                        mul = mul.value
                    total *= mul

                case TT.DIV:
                    div = element.factors[i+1].accept(self)
                    if isinstance(div, pf.ParsedLiteral):
                        div = div.value
                    total /= div

        return total

    def visitFactor(self, element: ast.FactorNode) -> int | float:
        """Visitor for a StringNode

        Parameters
        ----------
        element: StringNode
            The visited node

        Returns
        -------
        int | float
            The arithmetic value of the node
        """
        total = element.factors[0].accept(self)
        if isinstance(total, pf.ParsedLiteral):
            total = total.value

        match element.operation:
            case TT.PLUS:
                return pf.ParsedLiteral(total)

            case TT.MINUS:
                return pf.ParsedLiteral(-total)

            case TT.POW:
                for f in element.factors[1:]:
                    pow = f.accept(self)
                    if isinstance(pow, pf.ParsedLiteral):
                        pow = pow.value
                    total = total**pow
                return pf.ParsedLiteral(total)

    def visitStringExpr(self, element: ast.StringExprNode) -> str:
        """Visitor for a StringExprNode

        Parameters
        ----------
        element: StringExprNode
            The visited node

        Returns
        -------
        str
            The string value of the node
        """
        ret = ''
        for value in element.values:
            ret += str(value.accept(self))

        return pf.ParsedLiteral(ret)

    def visitVariableDefinition(self, element: ast.VariableDefinitionNode) -> str:
        """Visitor for a VariableDefinitionNode

        Parameters
        ----------
        element: VariableDefinitionNode
            The visited node

        Returns
        -------
        str
            The name of the declared variable

        Raises
        ------
        DSLParserVariableAlreadyUsed
            Tried to declare a varible that was already declared
        """
        if element.name in self.__variables:
            raise DSLParserVariableAlreadyUsed(element.name)

        self.__variables[element.name] = element.value.accept(self)

        return element.name

    def visitVariable(self, element: ast.VariableNode) -> object:
        """Visitor for a VariableNode

        Parameters
        ----------
        element: VariableNode
            The visited node

        Returns
        -------
        object
            The value of the declared variable

        Raises
        ------
        DSLParserVariableNotDeclared
            Tried to use a varible that was not declared
        """
        if element.name not in self.__variables:
            raise DSLParserVariableNotDeclared(element.name)

        return self.__variables[element.name]

    def visitInt(self, element: ast.IntNode) -> int:
        """Visitor for an IntNode

        Parameters
        ----------
        element: IntNode
            The visited node

        Returns
        -------
        int
            The value of the node
        """
        return int(element.value)

    def visitFloat(self, element: ast.FloatNode) -> float:
        """Visitor for a FloatNode

        Parameters
        ----------
        element: FloatNode
            The visited node

        Returns
        -------
        float
            The value of the node
        """
        return float(element.value)

    def visitBool(self, element: ast.BoolNode) -> bool:
        """Visitor for a BoolNode

        Parameters
        ----------
        element: BoolNode
            The visited node

        Returns
        -------
        bool
            The value of the node
        """
        return bool(element.value)

    def visitString(self, element: ast.StringNode) -> str:
        """Visitor for a BoolNode

        Parameters
        ----------
        element: BoolNode
            The visited node

        Returns
        -------
        bool
            The value of the node
        """
        return str(element.value.value)

    def visitIf(self, element: ast.IfNode) -> object | None:
        """Visitor for an IfNode

        Parameters
        ----------
        element: IfNode
            The visited node

        Returns
        -------
        object | None
            The value of the child node if the condition is true, else None
        """
        return element.ifCase.accept(self) if element.condition.accept(self).value\
            else None

    def visitIfElse(self, element: ast.IfElseNode):
        """Visitor for an IfElseNode

        Parameters
        ----------
        element: IfElseNode
            The visited node

        Returns
        -------
        object | None
            The value of the "if" child node if the condition is true, else the value\
                  of the "else" child node
        """
        return element.ifCase.accept(self) if element.condition.accept(self).value\
            else element.elseCase.accept(self)

    def visitAmount(self, element: ast.AmountNode) -> list[object]:
        """Visitor for an AmountNode

        Parameters
        ----------
        element: AmountNode
            The visited node

        Returns
        -------
        list[object]
            A list with the required number of object (resource, list item, ...)
        """
        var = element.variable.accept(self) if element.variable else None
        res = []
        amount = element.amount.accept(self).value
        if not isinstance(amount, int):
            raise DSLSyntaxException(amount.value.position)

        for _ in range(amount):
            res += element.node.accept(self)
            if var:
                self.__variables[var] += 1
        return res

    def visitParameter(self, element: ast.ParameterNode) -> pf.ParsedAttribute:
        """Visitor for an ParameterNode

        Parameters
        ----------
        element: ParameterNode
            The visited node

        Returns
        -------
        ParsedAttribute
            A ParsedAttribute object containing the name, position and value of \
                the node
        """
        return pf.ParsedAttribute(
            name=element.name.accept(self),
            position=element.position,
            value=element.value.accept(self),
        )

    def visitDict(self, element: ast.DictNode) -> pf.ParsedDict:
        """Visitor for an DictNode

        Parameters
        ----------
        element: DictNode
            The visited node

        Returns
        -------
        ParsedDict
            A ParsedDict object based on the node attributes
        """
        return pf.ParsedDict([v.accept(self) for v in element.value])

    def visitLiteral(self, element: ast.LiteralNode) -> pf.ParsedLiteral:
        """Visitor for an LiteralNode

        Parameters
        ----------
        element: LiteralNode
            The visited node

        Returns
        -------
        ParsedLiteral
            A ParsedLiteral object based on the node value
        """
        return pf.ParsedLiteral(element.value.accept(self))

    def visitList(self, element: ast.ListNode) -> pf.ParsedList:
        """Visitor for an ListNode

        Parameters
        ----------
        element: ListNode
            The visited node

        Returns
        -------
        ParsedList
            A ParsedLiteral object based on the node elements
        """
        return pf.ParsedList([v.accept(self) for v in element.value])

    def visitResource(self, element: ast.ResourceNode) -> list[pf.ParsedResource]:
        """Visitor for an ResourceNode

        Parameters
        ----------
        element: ResourceNode
            The visited node

        Returns
        -------
        int
            A ParsedResource object based on the node's attributes
        """
        return [
            pf.ParsedResource(
                type=element.type.accept(self).value,
                name=element.name.accept(self).value,
                position=element.position,
                attributes=[v.accept(self) for v in element.parameters.value],
            ),
        ]

    def visitFile(self, element: ast.FileNode):
        """Visitor for an FileNode

        Parameters
        ----------
        element: FileNode
            The visited node

        Returns
        -------
        ParsedFile
            A ParsedFile object based on the node's resources
        """
        file = pf.ParsedFile()
        for res in element.resources:
            file.resources += res.accept(self)

        return file



from thipster.engine.parsed_file import Position
from thipster.helpers import create_logger

from .exceptions import DSLParserNoEndingQuotes
from .lexer_position import LexerPosition
from .token import TOKENTYPES as TT
from .token import Token


class Lexer():
    """Lexer class for the DSL Parser
    """

    def __init__(self, files: dict[str, str]):
        """
        Parameters
        ----------
        files : dict[str, str]
            Dictionnary of files to tokenize, fileName : fileContent
        """
        self.__logger = create_logger(__name__)
        self.__files = files
        self.__tokenList = []
        self.__lexerPosition = LexerPosition()
        self.__currentChar = ''

    @property
    def files(self):
        return self.__files

    @property
    def tokenList(self):
        return self.__tokenList

    def run(self) -> list[Token]:
        """Launch the Lexer

        Returns
        -------
        list[Token]
            List of Tokens representing the input files
        """
        self.__logger.info('Start Lexer')
        for fileName, fileContent in self.files.items():
            self.lex(fileContent, fileName, 1, 1)

        self.__logger.info(
            'Shuting down Lexer, processed %d file(s) and generated %d tokens',
            len(self.__files),
            len(self.__tokenList),
        )
        return self.__tokenList

    def lex(
        self,
        codeString: str,
        fileName: str,
        startLine: int = 1,
        startColumn: int = 1,
    ) -> None:
        """Tokenize a file contents

        Parameters
        ----------
        codeString : str
            Content of the file
        fileName : str
            Name of the file
        startLine : int, optional
            Starting line of the content in the file, by default 1
        startColumn : int, optional
            Starting column of the content in the file, by default 1

        Raises
        ------
        DSLParserNoEndingQuotes
            Syntax error : detected a starting quote but no ending one
        """
        self.__logger.debug('Lex file %s', fileName)
        self.__lexerPosition = LexerPosition(fileName, startLine, startColumn)

        for char in codeString:
            self.__logger.debug('char %s', char)
            self.__currentChar = char
            if not self.__lexerPosition.isQuotedString:
                self.__handleSyntaxTokens()
                continue

            if char in ['"', "'"]:
                self.__handleDoubleQuotes(char)()
                continue

            if char == '\n':
                if self.__lexerPosition.isCurrentTokenMultiLine:
                    self.__lexerPosition.newLine()
                    continue
                position = str(
                    Position(
                        self.__lexerPosition.currentFile,
                        self.__lexerPosition.currentLine,
                        self.__lexerPosition.currentColumn,
                    ),
                )
                self.__logger.error(
                    f'Invalid syntax, missing ending quotes at : {position}',
                )
                raise DSLParserNoEndingQuotes(position)

            self.__iterateNextChar()

        self.__addFileEnd()
        self.__logger.debug('Finished lexing file %s', fileName)

    def __iterateNextChar(self):
        """Iterate to the next character

        Add the current character to the current Token
        """
        self.__lexerPosition.addCharToCurrentToken(self.__currentChar)
        self.__lexerPosition.nextColumn()

    def __handleEmptyToken(self):
        pass

    def __handleSyntaxTokens(self) -> None:
        """Handle single character tokens

        Check if the current character is a special token and calls the corresponding
        function.
        If it isn't, it calls the '__iterateNextChar' function
        """
        singleCharTokens = {
            ':': self.__handleColonToken,
            ',': self.__handleCommaToken,
            '[': self.__handleBracketsStartToken,
            ']': self.__handleBracketsEndToken,
            '(': self.__handleParenthesesStartToken,
            ')': self.__handleParenthesesEndToken,
            '"': self.__handleDoubleQuotes('"'),
            "'": self.__handleDoubleQuotes("'"),
            '#': self.__handleHashToken,
            '\n': self.__handleNewlineToken,
            '\\': self.__handleBackSlashToken,
            '\t': self.__handleTabToken,
            ' ': self.__handleWhitespace,

            '=': self._handleEQToken,
            '/': self._handleDivToken,
            '-': self._handleMinusToken,
            '+': self._handlePlusToken,
            '*': self._handleMulToken,
            '^': self._handlePowToken,
            '<': self._handleLTToken,
            '>': self._handleGTToken,
            '!': self._handleExclamationToken,
        }

        singleCharTokens.get(self.__currentChar, self.__iterateNextChar)()

    def __handleCurrentToken(self) -> None:
        """Process the current stored token

        Check if the current token is a keyword, calls the cprresponding function, or
        the '__handleLiteralsAndVariables' otherwise.
        """
        currentTokenLower = self.__lexerPosition.currentToken.lower()

        keyWordsAndTokens = {
            '': self.__handleEmptyToken,
            'amount': self.__handleAmountToken,
            'and': self.__handleAndToken,
            'if': self.__handleIfToken,
            'elif': self.__handleElifToken,
            'else': self.__handleElseToken,
            'or': self.__handleOrToken,
            'true': self.__handleBooleanToken,
            'false': self.__handleBooleanToken,
            'not': self.__handleNotToken,
        }

        keyWordsAndTokens.get(
            currentTokenLower, self.__handleLiteralsAndVariables,
        )()

    def __handleLiteralsAndVariables(self) -> None:
        """Create a literal or var token

        Checks the current token to create a var, int, float or string token
        """
        currentToken = self.__lexerPosition.currentToken
        if self.__lexerPosition.isCurrentTokenAVariable:
            self.__addLiteralTokenToList(TT.VAR, currentToken)
            self.__lexerPosition.setIsVariable(False)

        elif currentToken.isdigit():
            self.__addLiteralTokenToList(TT.INT, currentToken)

        elif self.__isfloat(currentToken):
            self.__addLiteralTokenToList(TT.FLOAT, currentToken)

        else:
            self.__addLiteralTokenToList(TT.STRING, currentToken)

    def __addLiteralTokenToList(self, tokenType: str, value: str):
        """Add a literal token to the token list

        Parameters
        ----------
        tokenType : str
            Type of the token
        value : str
            Value of the token
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(
            tokenType,
            value,
            isCurrentToken=True,
        )

    def __addBaseToken(
        self,
        tokenType: str,
        value: str | None = None,
        isCurrentToken: bool = False,
    ) -> None:
        """Add a token to the token list

        Parameters
        ----------
        tokenType : str
            Type of the token
        value : str, optional
            Value of the token, by default None
        isCurrentToken : bool, optional
            Indicates if this token is the current stored token, by default False
        """
        if not isCurrentToken:
            self.__addTokenToList(
                Token(
                    self.__lexerPosition.currentCharPosition,
                    tokenType,
                    value,
                ),
            )
            return

        self.__addTokenToList(
            Token(
                self.__lexerPosition.currentTokenPosition,
                tokenType,
                value,
            ),
        )

    def __basePositionUpdate(
        self,
        resetCurrentToken: bool = True,
    ) -> None:
        """Update the lexer's current position

        Parameters
        ----------
        resetCurrentToken : bool, optional
            Indicates if the current stored token should be reset, by default True
        """
        self.__lexerPosition.nextColumn()
        if resetCurrentToken:
            self.__lexerPosition.resetCurrentToken()

        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__lexerPosition.setIsVariable(False)

    def __handleBaseToken(
        self,
        tokenType: str,
        value: str | None = None,
        isCurrentToken: bool = False,
        handleCurrentToken: bool = True,
        resetCurrentToken: bool = True,
    ) -> None:
        """Handle a token creation and lexer position

        Parameters
        ----------
        tokenType : str
            Type of the token
        value : str, optional
            Value of the token, by default None
        isCurrentToken : bool, optional
            Indicates if this token is the current stored token, by default False
        handleCurrentToken : bool, optional
            Indicates if the current stored token should be processed, by default True
        resetCurrentToken : bool, optional
            Indicates if the current stored token should be reset, by default True
        """
        if handleCurrentToken:
            self.__handleCurrentToken()
        self.__addBaseToken(tokenType, value, isCurrentToken)
        self.__basePositionUpdate(resetCurrentToken)

    def __handleColonToken(self):
        """Handle a COLON token ':'
        """
        self.__handleBaseToken(TT.COLON)

    def __handleCommaToken(self):
        """Handle a COMMA token ','
        """
        self.__handleBaseToken(TT.COMMA)

    def __handleBracketsStartToken(self):
        """Handle a BRACKETS_START token '['
        """
        self.__handleBaseToken(TT.BRACKETS_START)

    def __handleBracketsEndToken(self):
        """Handle a BRACKETS_END token ']'
        """
        self.__handleBaseToken(TT.BRACKETS_END)

    def __handleParenthesesStartToken(self):
        """Handle a PARENTESES_START token '('
        """
        self.__handleBaseToken(TT.PARENTHESES_START)

    def __handleParenthesesEndToken(self):
        """Handle a PARENTHESES_END token ')'
        """
        self.__handleBaseToken(TT.PARENTHESES_END)

    def _handlePlusToken(self):
        """Handle a PLUS token '+'
        """
        self.__handleBaseToken(TT.PLUS)

    def _handleMinusToken(self):
        """Handle a MINUS token '-'
        """
        self.__handleBaseToken(TT.MINUS)

    def _handleMulToken(self):
        """Handle a MUL token '*'
        """
        self.__handleBaseToken(TT.MUL)

    def _handleDivToken(self):
        """Handle a DIV token '/'
        """
        self.__handleBaseToken(TT.DIV)

    def _handleEQToken(self):
        """Handle a EQ token '='
        """
        self.__handleBaseToken(TT.EQ)

    def _handleExclamationToken(self):
        """Handle a NE token '!='
        """
        self.__handleBaseToken(TT.EXCLAMATION)

    def _handleLTToken(self):
        """Handle a LT token '<'
        """
        self.__handleBaseToken(TT.LT)

    def _handleGTToken(self):
        """Handle a GT token '>'
        """
        self.__handleBaseToken(TT.GT)

    def _handlePowToken(self):
        """Handle a POW token '^'
        """
        self.__handleBaseToken(TT.POW)

    def __handleHashToken(self) -> None:
        """Handle a HASH token '#'

        Sets a variable to indicate that the following literal is a Variable
        """
        self.__handleCurrentToken()
        self.__basePositionUpdate()
        self.__lexerPosition.setIsVariable(True)

    def __handleWhitespace(self) -> None:
        """Handle a WHITESPACE token ' '

        Replaces 4 following whitespaces by a TAB token
        """
        self.__handleCurrentToken()
        self.__addBaseToken(TT.WHITESPACE)
        self.__lexerPosition.nextColumn()
        self.__lexerPosition.resetCurrentToken()
        self.__lexerPosition.setIsVariable(False)
        self.__lexerPosition.addConsecutiveWhitespace()
        if self.__lexerPosition.consecutiveWhitespaces == 4:
            self.__lexerPosition.nextColumn(-4)
            self.__rmLastTokens(4)
            self.__addBaseToken(TT.TAB)
            self.__lexerPosition.nextColumn(4)
            self.__lexerPosition.resetConsecutiveWhitespaces()
            self.__lexerPosition.setCurrentTokenIndex()

    def __handleNewlineToken(self) -> None:
        """Handle a NEWLINE token '\\n'
        """
        if not self.__lexerPosition.isCurrentTokenMultiLine:
            self.__handleBaseToken(TT.NEWLINE)

        self.__lexerPosition.newLine()
        if not self.__lexerPosition.isCurrentTokenMultiLine:
            self.__lexerPosition.setCurrentTokenIndex()

        self.__lexerPosition.setIsMultiLine(value=False)

    def __handleBackSlashToken(self) -> None:
        """Handle a BACKSLASH token '\\'

        Sets a variable to indicate that the following line is part of the same token.
        """
        self.__lexerPosition.setIsMultiLine(value=True)
        self.__lexerPosition.nextColumn()

    def __handleTabToken(self) -> None:
        """Handle a TAB token '\\t'
        """
        self.__handleBaseToken(TT.TAB)

    def __handleDoubleQuotes(self, char) -> None:
        """Handle a DOUBLEQUOTES token '"'

        Sets a variable to indicate that the following characters are a STRING token.
        """

        def quoteHandler(self=self):

            if self.__lexerPosition.isQuotedString:
                if char == self.__lexerPosition.isQuotedString:
                    self.__addLiteralTokenToList(
                        TT.STRING, self.__lexerPosition.currentToken,
                    )
                    self.__lexerPosition.isQuotedString = False
                    self.__basePositionUpdate()
                else:
                    self.__iterateNextChar()
            else:
                self.__handleCurrentToken()
                self.__lexerPosition.nextColumn()
                self.__lexerPosition.isQuotedString = char
                self.__lexerPosition.resetCurrentToken()

        return quoteHandler

    def __handleAmountToken(self):
        """Handle an AMOUNT token 'amount'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(TT.AMOUNT, isCurrentToken=True)

    def __handleAndToken(self):
        """Handle an AND token 'and'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(TT.AND, isCurrentToken=True)

    def __handleIfToken(self):
        """Handle an IF token 'if'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(TT.IF, isCurrentToken=True)

    def __handleElifToken(self):
        """Handle an ELIF token 'elif'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(TT.ELIF, isCurrentToken=True)

    def __handleElseToken(self):
        """Handle an ELSE token 'else'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(TT.ELSE, isCurrentToken=True)

    def __handleNotToken(self):
        """Handle an OR token 'or'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(TT.NOT, isCurrentToken=True)

    def __handleOrToken(self):
        """Handle an OR token 'or'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(TT.OR, isCurrentToken=True)

    def __handleBooleanToken(self):
        """Handle a BOOLEAN token 'true' or 'false'
        """
        self.__lexerPosition.resetConsecutiveWhitespaces()
        self.__addBaseToken(
            TT.BOOLEAN, self.__lexerPosition.currentToken, True,
        )

    def __handleEofToken(self) -> None:
        """Add an EOF token at the end of each file
        """
        self.__addBaseToken(TT.EOF)

    def __addTokenToList(self, token: Token | None) -> None:
        """Add a new token to the parser's token list

        Parameters
        ----------
        token : Token | None
            New token to add
        """
        if not token:
            pass
        self.__tokenList.append(token)

    def __rmLastTokens(self, amount: int = 1):
        self.__tokenList = self.__tokenList[:-amount]

    def __addFileEnd(self) -> None:
        """Add a NEWLINE and an EOF token at the end of each file
        """
        if len(self.__lexerPosition.currentToken.strip()) > 0:
            self.__handleCurrentToken()
            self.__lexerPosition.resetCurrentToken()
        self.__handleNewlineToken()
        self.__handleEofToken()

    def __isfloat(self, num: str) -> bool:
        """Check if the string is a float

        Parameters
        ----------
        num : str
            The input string to check

        Returns
        -------
        bool
            Is the string a float true or false
        """
        try:
            float(num)
            return True
        except ValueError:
            return False



import os

from thipster.engine import I_Parser
from thipster.engine.parsed_file import ParsedFile
from .dsl_parser import DSLParser
from .yaml_parser import YAMLParser


class ParserPathNotFound(Exception):
    def __init__(self, path, *args: object) -> None:
        super().__init__(*args)

        self.__message = f'Path not found : {path}'

    @property
    def message(self):
        return self.__message


class ParserFactory(I_Parser):

    __parsers = {
        '.yaml': YAMLParser,
        '.yml': YAMLParser,
        '.jinja': YAMLParser,
        '.thips': DSLParser,
    }

    def addParser(parser: I_Parser, extensions: list[str]):
        ParserFactory.__parsers.update({e: parser for e in extensions})

    def __getfiles(self, path: str) -> list[str]:
        """Recursively get all files names in the requested directory and its\
              sudirectories
        Can be run on a path file aswell

        Parameters
        ----------
        path: str
            Path to run this function into

        Returns
        -------
        list[str]
            A list of all the filenames
        """

        path = os.path.abspath(path)

        if not os.path.exists(path):
            raise ParserPathNotFound(path)

        files = []

        if os.path.isdir(path):
            for content in os.listdir(path):
                files += self.__getfiles(f'{path}/{content}')

        if os.path.isfile(path):
            return [path]

        return files

    def __noParser(self):
        raise Exception()

    def run(self, path: str) -> ParsedFile:
        """Run the ParserFactory

        Parameters
        ----------
        path: str
            Path to run the parser into

        Returns
        -------
        ParsedFile
            A ParsedFile object with the content of all the files in the input path
        """
        files = self.__getfiles(path)

        res = ParsedFile()
        for file in files:
            parsedFile = self.__getParser(file).run(file)
            res.resources += parsedFile.resources

        return res

    def __getParser(self, path) -> I_Parser:

        _, pathExtension = os.path.splitext(path)

        return ParserFactory.__parsers.get(pathExtension, self.__noParser)



"""parsedFile.py Module
"""

from abc import ABC, abstractmethod


class I_ParsedValue(ABC):
    """Parsed Value Interface
    """

    @property
    @abstractmethod
    def value(self):
        """Abstract value attribute
        """

        pass


class Position():
    """Class representing the position of a token

    Indicates the initial position of a token, resource or character in the input files.
    It includes the file name, line and column numbers of the designated element.
    """

    def __init__(self, fileName: str, ln: int, col: int):
        """
        Parameters
        ----------
        fileName : str
            File name
        ln : int
            Line number
        col : int
            Column number
        """

        self.__fileName = fileName
        self.__ln = ln
        self.__col = col

    @property
    def fileName(self):
        """Name of the file
        """

        return self.__fileName

    @property
    def ln(self):
        """Line number of the token
        """

        return self.__ln

    @property
    def col(self):
        """Column number of the token
        """

        return self.__col

    def __str__(self) -> str:
        """Returns a string of the position object

        Returns
        -------
        str
            Displays the filename, line and column
        """

        return f'(File : {self.fileName}, Ln {self.ln}, Col {self.col})'

    def __eq__(self, __value: object) -> bool:
        """Check if 2 positions are equal

        Parameters
        ----------
        __value : object
            Position to compare

        Returns
        -------
        bool
            True if both positions are equal or False otherwise

        Raises
        ------
        TypeError
            If '__value' is not a Position
        """

        if isinstance(__value, Position):
            return (
                self.fileName == __value.fileName and
                self.ln == __value.ln and
                self.col == __value.col
            )
        else:
            raise TypeError('Value must be a Position')


class ParsedAttribute():
    """Class reprensenting a Parsed Attribute Object
    """

    def __init__(self, name: str, position: Position | None, value: I_ParsedValue):
        """
        Parameters
        ----------
        name : str
        position : Position
        value : I_parsed_Value
        """

        self.__name = name
        self.__position = position
        self.__value = value

    @property
    def value(self):
        """Value of the parsed attribute
        """

        return self.__value.value

    @property
    def name(self):
        """Name of the parsed attribute
        """

        return self.__name

    @property
    def position(self):
        """Position of the parsed attribute
        """

        return self.__position


class ParsedList(I_ParsedValue):
    """Class representing a Parsed List Object
    """

    def __init__(self, value: list[I_ParsedValue]):
        """
        Parameters
        ----------
        value : list[I_Parsed_Value]
        """

        super().__init__()
        self.__value = value

    @property
    def value(self):
        """Value of the parsed list
        """

        return self.__value

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if self.i > len(self.__value):
            raise StopIteration
        else:
            ret = self.value[self.i]
            self.i += 1

        return ret


class ParsedLiteral(I_ParsedValue):
    """Class representing a Parsed Literal Object
    """

    def __init__(self, value: bool | int | float | str):
        """
        Parameters
        ----------
        value : Literal
        """

        super().__init__()
        self.__value = value

    @property
    def value(self):
        """Value of the Parsed literal
        """

        return self.__value


class ParsedDict(I_ParsedValue):
    """Class representing a Parsed Dictionnary Object
    """

    def __init__(self, value: list[ParsedAttribute]):
        """
        Parameters
        ----------
        value : list[Parsed_Attribute]
        """

        super().__init__()
        self.__value = value

    @property
    def value(self):
        """Value of the Parsed Dictionary
        """

        return self.__value


class ParsedResource():
    """Class representing a Parsed Resource
    """

    def __init__(
            self,
            type: str,
            name: str,
            position: Position | None,
            attributes: list[ParsedAttribute],
    ):
        """
        Parameters
        ----------
        type : str
        name : str
        position : Position
        attributes : list[Parsed_Attribute]
        """

        self.__type = type
        self.__name = name
        self.__position = position
        self.__attributes = attributes

    @property
    def type(self) -> str:
        """Type of the Parsed Resource
        """

        return self.__type

    @property
    def name(self) -> str:
        """Name of the Parsed Resource
        """

        return self.__name

    @property
    def position(self) -> Position | None:
        """Position of the Parsed Resource in its origin file
        """

        return self.__position

    @property
    def attributes(self) -> list[ParsedAttribute]:
        """List of Attributes of the Parsed Resource
        """

        return self.__attributes


class ParsedFile():
    """Class representing a Parsed File

    Object containing a list of parsed resources making up a file.

    """

    def __init__(self):
        self.__resources = []

    @property
    def resources(self) -> list[ParsedResource]:
        """List of the resources defined in the Parsed File
        """

        return self.__resources

    @resources.setter
    def resources(self, resources):
        """Set the resources list
        """

        self.__resources = resources



from abc import ABC, abstractmethod

from thipster.engine.parsed_file import ParsedFile


class I_Parser(ABC):
    """Parser module interface
    """
    @abstractmethod
    def run(self, path: str) -> ParsedFile:
        """Abstract run method

        Parameters
        ----------
        path : str
            The path of the filesor directory to be processed

        Returns
        -------
        ParsedFile
            Object containing a representation of an input file broken down into
            individual resources

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        raise NotImplementedError('Should implement run()')



import os

from .json import JSONRepo


class LocalRepo(JSONRepo):
    def __init__(self, path: str) -> None:
        super().__init__()
        self.__repo = path

    def get_json(self, name: str) -> str | bytes | bytearray:
        name += '.json'
        with open(os.path.join(self.__repo, name), 'r') as file:
            return file.read()



"""Central engine module for THipster.

This module contains the Engine class, which is the central class of the THipster
package as well as the interfaces for the various components of the application.
"""

from .engine import Engine
from .i_auth import I_Auth
from .i_parser import I_Parser
from .i_repository import I_Repository
from .i_terraform import I_Terraform



import thipster.parser.dsl_parser.ast as ast

from .exceptions import (
    DSLConditionException,
    DSLSyntaxException,
    DSLUnexpectedEOF,
)
from .token import TOKENTYPES as TT
from .token import Token


class TokenParser():
    def __init__(self, tokens: list[Token]) -> None:
        self.__tokens = tokens

    def run(self) -> ast.FileNode:
        self.__rm_empty_lines()
        tree = ast.FileNode()
        try:
            self.__trim_newlines()
            while self.__get_next_type() != TT.EOF:
                self.__trim_newlines()

                tree.add(self.__create_resource())
                self.__trim_newlines()
        except Exception as e:
            raise e

        return tree

    def __next(self, expected: TT | list[TT] | None = None) -> Token:
        """Get next token and pop it from the list"""
        tok = self.__get_next_type()

        if expected:
            if type(expected) is list:
                if tok not in expected:
                    raise DSLSyntaxException(self.__tokens[0], expected)
            elif tok != expected:
                raise DSLSyntaxException(self.__tokens[0], expected)
        return self.__tokens.pop(0)

    def __check(self, expected: TT, index: int = 0) -> Token | None:
        """Check if the type of the next token is equal to the expected parameter. \
        Pop it from the list in that case"""
        token = self.__get_next_type(index=index)
        if token != expected:
            return None

        return self.__next(expected)

    def __get_next_type(self, index: int = 0):
        """Get the type of the next token"""
        if len(self.__tokens) <= index:
            raise DSLUnexpectedEOF

        return self.__tokens[index].tokenType

    def __trim_newlines(self):
        while self.__check(TT.NEWLINE):
            pass

    def __rm_empty_lines(self):
        # Detect empty line
        empty_types = [TT.TAB]
        end = 0
        while end < len(self.__tokens):
            begin = end
            while self.__get_next_type(end) in empty_types:
                end += 1

            if self.__get_next_type(index=end) == TT.NEWLINE:
                for _ in range(begin, end):
                    self.__tokens.pop(begin)
            end += 1

    def __get_newline(self):
        nl = self.__next(TT.NEWLINE)
        self.__trim_newlines()

        return nl

    def __get_whitespaces(self):
        while self.__check(TT.WHITESPACE):
            pass

    def __get_tabs(self, indent: int) -> bool:
        """Check if the number of tabs is correct/ if it is the end of the block"""

        if self.__get_next_type() == TT.EOF:
            return False
        elif self.__get_next_type(indent-1) == TT.TAB:
            for i in range(indent):
                self.__next(TT.TAB)
            return True
        return False

    def __create_resource(self, indent=0) \
            -> ast.ResourceNode | ast.IfNode | ast.AmountNode:
        """type, name, ":", [amt_ctrl], [if_ctrl] ,"\\n"
                                (list | dict | {parameter, "\\n"})"""
        try:
            for _ in range(indent):
                self.__next(TT.TAB)

            resType = self.__get_type()
            self.__get_whitespaces()
            name = self.__get_string_expr()
            self.__get_whitespaces()
            self.__next(TT.COLON)
            self.__get_whitespaces()

            nbCtrl = self.__get_nb_ctrl()
            self.__get_whitespaces()
            ifCtrl = self.__get_if_ctrl()
            self.__get_whitespaces()

            self.__get_newline()

            properties = self.__get_properties(indent+1)

        except DSLSyntaxException as e:
            raise e

        resource = ast.ResourceNode(
            resourceType=resType,
            name=name,
            parameters=properties,
        )

        if ifCtrl:
            ifCtrl.ifCase = resource
            resource = ifCtrl

        if nbCtrl:
            nbCtrl.node = resource
            resource = nbCtrl

        return resource

    def __get_parameter(self, indent: int) -> ast.ParameterNode:
        """name, ":", (value, [if_else_ctrl] | [if_ctrl], "\\n", (list | dict))"""
        try:
            name = self.__next(TT.STRING)
            self.__get_whitespaces()
            self.__next(TT.COLON)
            self.__get_whitespaces()
        except DSLSyntaxException as e:
            raise e

        next = self.__get_next_type()

        if next not in [
            TT.IF,
            TT.NEWLINE,
        ]:
            # value, [if_else_ctrl]
            try:
                value = self.__get_value()
                self.__get_whitespaces()
                ifElseCtrl = self.__get_if_else_ctrl()
                self.__get_whitespaces()
            except:
                raise

            parameter = ast.ParameterNode(
                name=ast.StringNode(name),
                value=value,
            )

            if ifElseCtrl:
                ifElseCtrl.ifCase = parameter.value
                parameter.value = ifElseCtrl

            return parameter

        # [if_ctrl], "\n" (liste | dict)
        try:
            ifCtrl = self.__get_if_ctrl()
            self.__get_whitespaces()
            nl = self.__get_newline()
            properties = self.__get_properties(indent+1)
        except DSLSyntaxException as e:
            if e.tok.tokenType == TT.TAB:
                raise DSLSyntaxException(
                    token=nl,
                    expected=TT.STRING,
                )

            raise e

        parameter = ast.ParameterNode(
            name=ast.StringNode(name),
            value=properties,
        )

        if ifCtrl:
            ifCtrl.ifCase = parameter
            parameter = ifCtrl

        return parameter

    def __get_properties(self, indent: int) -> list[ast.ParameterNode]:
        """(list | dict)
        """
        i = indent
        try:
            next = self.__get_next_type(indent)
            while next == TT.WHITESPACE:
                i += 1
                next = self.__get_next_type(i)

            if next == TT.STRING:
                props = self.__get_dict(indent)

            elif next == TT.MINUS:
                props = self.__get_list(indent)

            else:
                raise DSLSyntaxException(self.__next(), TT.TAB)

        except DSLUnexpectedEOF as eof:
            if indent > 1:
                raise DSLUnexpectedEOF from eof

            if eof.__cause__:
                raise DSLUnexpectedEOF from eof.__cause__

            props = ast.DictNode([])

        return props

    def __get_list(self, indent: int) -> ast.ListNode:
        """{ "-", value, [amt_ctrl], [if_else_ctrl], "\\n"}"""
        items = []

        try:
            while self.__get_tabs(indent):
                self.__check(TT.MINUS)
                self.__get_whitespaces()

                value = self.__get_value()
                self.__get_whitespaces()

                amountCtrl = self.__get_nb_ctrl()
                self.__get_whitespaces()

                ifElseCtrl = self.__get_if_else_ctrl()
                self.__get_whitespaces()

                if ifElseCtrl:
                    ifElseCtrl.ifCase = value
                    value = ifElseCtrl

                if amountCtrl:
                    amountCtrl.node = value
                    value = amountCtrl

                items.append(value)

                self.__get_newline()
        except Exception as e:
            raise e

        self.__tokens.insert(
            0, Token(
                position=self.__tokens[0].position,
                tokenType=TT.NEWLINE,
            ),
        )

        return ast.ListNode(items)

    def __get_inline_list(self) -> ast.ListNode:
        items = []

        self.__next(TT.BRACKETS_START)
        self.__get_whitespaces()
        next = self.__get_next_type()

        if next in [
            TT.BOOLEAN,
            TT.FLOAT,
            TT.INT,
            TT.STRING,
            TT.VAR,
            TT.BRACKETS_START,
        ]:
            items.append(self.__get_value())
            while not self.__check(TT.BRACKETS_END):
                self.__next([TT.COMMA, TT.NEWLINE])
                while self.__check(TT.TAB):
                    pass
                self.__get_whitespaces()

                items.append(self.__get_value())
                self.__get_whitespaces()
        else:
            self.__next(TT.BRACKETS_END)

        return ast.ListNode(items)

    def __get_dict(self, indent: int) -> ast.DictNode:
        """{ parameter, "\\n" }
        """
        props = []

        try:
            while self.__get_tabs(indent):
                self.__get_whitespaces()
                props.append(self.__get_parameter(indent))
                self.__get_whitespaces()
                self.__get_newline()
        except Exception as e:
            raise e

        self.__tokens.insert(
            0, Token(
                position=self.__tokens[0].position,
                tokenType=TT.NEWLINE,
            ),
        )

        return ast.DictNode(props)

    def __get_nb_ctrl(self) -> ast.AmountNode | None:
        """"amount", ":", int, ["#", var]"""

        try:
            nextToken = self.__check(TT.AMOUNT)
            if not nextToken:
                return None
            self.__get_whitespaces()

            self.__next(TT.COLON)
            self.__get_whitespaces()
            nb = self.__get_arith_expr()
            self.__get_whitespaces()
        except DSLSyntaxException as e:
            raise e

        var = self.__check(TT.VAR)

        return ast.AmountNode(
            amount=nb,
            variable=ast.VariableDefinitionNode(
                var, ast.IntNode(Token(None, TT.INT, 1)),
            )
            if var else None,
            node=None,
        )

    def __get_if_ctrl(self) -> ast.IfNode | None:
        """"if", condition"""

        condition = self.__check(TT.IF)
        if not condition:
            return None
        self.__get_whitespaces()

        try:
            condition = self.__get_comp_expr()
            self.__get_whitespaces()
        except DSLSyntaxException as e:
            raise e

        return ast.IfNode(
            condition=condition,
            ifCase=None,
        )

    def __get_if_else_ctrl(self) -> ast.IfElseNode | None:
        """if_ctrl, ["else" , valeur]"""
        ifCtrl = self.__get_if_ctrl()
        self.__get_whitespaces()
        if not ifCtrl:
            return None

        if self.__check(TT.ELSE):
            self.__get_whitespaces()
            elseCase = self.__get_value()
            self.__get_whitespaces()
        else:
            elseCase = None

        return ast.IfElseNode(
            condition=ifCtrl.condition,
            ifCase=None,
            elseCase=elseCase,
        )

    def __get_value(self) -> ast.LiteralNode:
        nextType = self.__get_next_type()
        match nextType:
            case TT.STRING:
                return self.__get_string_expr()

            case TT.VAR:
                return ast.LiteralNode(ast.VariableNode(self.__next(TT.VAR)))

            case TT.BRACKETS_START:
                return self.__get_inline_list()

            case TT.NOT | TT.BOOLEAN:
                return self.__get_comp_expr()

            case _:
                value = self.__get_arith_expr()
                self.__get_whitespaces()

                nextType = self.__get_next_type()
                if nextType in [TT.EE, TT.LT, TT.GT, TT.LTE, TT.GTE]:
                    op = self.__next([TT.EE, TT.LT, TT.GT, TT.LTE, TT.GTE])
                    self.__get_whitespaces()
                    expr2 = self.__get_comp_expr()
                    self.__get_whitespaces()
                    self.__next(TT.PARENTHESES_END)
                    self.__get_whitespaces()

                    return ast.CompExprNode(value, op, expr2)
                return value

    def __get_comp_expr(self) -> ast.CompExprNode:
        try:
            nextType = self.__get_next_type()
            match nextType:

                case TT.NOT:
                    op = self.__next(TT.NOT)
                    self.__get_whitespaces()
                    expr = self.__get_comp_expr()
                    self.__get_whitespaces()
                    return ast.CompExprNode(expr, op)

                case TT.PARENTHESES_START:
                    self.__next(TT.PARENTHESES_START)
                    self.__get_whitespaces()
                    expr1 = self.__get_comp_expr()
                    self.__get_whitespaces()
                    op = self.__next([TT.OR, TT.AND])
                    self.__get_whitespaces()
                    expr2 = self.__get_comp_expr()
                    self.__get_whitespaces()
                    self.__next(TT.PARENTHESES_END)
                    self.__get_whitespaces()
                    return ast.CompExprNode(expr1, op, expr2)

                case TT.BOOLEAN:
                    return ast.LiteralNode(ast.BoolNode(self.__next(TT.BOOLEAN)))

                case _:
                    expr1 = self.__get_arith_expr()
                    self.__get_whitespaces()
                    op = self.__next([TT.EQ, TT.LT, TT.GT, TT.EXCLAMATION])

                    # ==
                    if op.tokenType is TT.EQ:
                        self.__next(TT.EQ)
                        op = Token(op.position, TT.EE)

                    # !=
                    if op.tokenType is TT.EXCLAMATION:
                        self.__next(TT.EQ)
                        op = Token(op.position, TT.NE)

                    nextType = self.__get_next_type()
                    if nextType is TT.EQ:
                        self.__next(TT.EQ)
                        match op.tokenType:
                            case TT.LT:
                                op = Token(op.position, TT.LTE)
                            case TT.GT:
                                op = Token(op.position, TT.GTE)

                    self.__get_whitespaces()
                    expr2 = self.__get_arith_expr()
                    self.__get_whitespaces()
                    return ast.CompExprNode(expr1, op, expr2)

        except DSLSyntaxException as e:
            raise DSLConditionException(e.tok)

    def __get_arith_expr(self) -> ast.ArithExprNode:
        terms = []
        op = []
        terms.append(self.__get_term())
        self.__get_whitespaces()

        nextType = self.__get_next_type()
        while nextType in [TT.PLUS, TT.MINUS]:
            op.append(nextType)
            terms.append(self.__get_term())
            self.__get_whitespaces()
            nextType = self.__get_next_type()

        return ast.ArithExprNode(terms, op)

    def __get_term(self) -> ast.TermNode:
        factors = []
        op = []
        factors.append(self.__get_factor())
        self.__get_whitespaces()

        nextType = self.__get_next_type()
        while nextType in [TT.MUL, TT.DIV]:
            op.append(self.__next([TT.MUL, TT.DIV]).tokenType)
            self.__get_whitespaces()
            factors.append(self.__get_factor())
            self.__get_whitespaces()
            nextType = self.__get_next_type()

        return ast.TermNode(factors, op)

    def __get_factor(self) -> ast.FactorNode:
        nextType = self.__get_next_type()

        match nextType:
            case TT.PLUS | TT.MINUS:
                op = self.__next([TT.PLUS, TT.MINUS]).tokenType
                self.__get_whitespaces()
                factor = self.__get_factor()
                self.__get_whitespaces()
                return ast.FactorNode([factor], op)

            case _:
                factors = [self.__get_atom()]
                self.__get_whitespaces()

                nextType = self.__get_next_type()
                op = TT.PLUS
                while nextType == TT.POW:
                    op = self.__next(TT.POW).tokenType
                    self.__get_whitespaces()
                    factors.append(self.__get_factor())
                    self.__get_whitespaces()
                    nextType = self.__get_next_type()
                return ast.FactorNode(factors, op)

    def __get_atom(self):
        nextType = self.__get_next_type()
        match nextType:
            case TT.INT:
                return ast.IntNode(self.__next())

            case TT.FLOAT:
                return ast.FloatNode(self.__next())

            case TT.PARENTHESES_START:
                self.__next(TT.PARENTHESES_START)
                self.__get_whitespaces()
                expr = self.__get_arith_expr()
                self.__get_whitespaces()
                self.__next(TT.PARENTHESES_END)
                self.__get_whitespaces()
                return expr

            case _:
                raise DSLSyntaxException(
                    self.__tokens[0], [TT.INT, TT.FLOAT, TT.PARENTHESES_START],
                )

    def __get_string_expr(self):
        values = []
        token = self.__next([TT.STRING, TT.VAR])
        match token.tokenType:
            case TT.STRING:
                values.append(ast.StringNode(token))

            case TT.VAR:
                values.append(ast.VariableNode(token))

        nextType = self.__get_next_type()
        while nextType in [TT.STRING, TT.VAR, TT.INT, TT.MINUS, TT.DIV]:
            match nextType:
                case TT.STRING:
                    values.append(ast.StringNode(self.__next(TT.STRING)))

                case TT.VAR:
                    values.append(ast.VariableNode(self.__next(TT.VAR)))

                case TT.INT:
                    values.append(ast.IntNode(self.__next(TT.INT)))

                case TT.DIV:
                    token = Token(self.__next(TT.DIV).position, TT.STRING, '/')
                    values.append(ast.StringNode(token))

                case TT.MINUS:
                    token = Token(
                        self.__next(
                            TT.MINUS,
                        ).position, TT.STRING, '-',
                    )
                    values.append(ast.StringNode(token))
            nextType = self.__get_next_type()

        return ast.StringExprNode(values)

    def __get_type(self):
        values = []
        values.append(ast.StringNode(self.__next(TT.STRING)))

        nextType = self.__get_next_type()
        while nextType in [TT.STRING, TT.MINUS, TT.DIV]:
            match nextType:
                case TT.STRING:
                    values.append(ast.StringNode(self.__next(TT.STRING)))

                case TT.DIV:
                    token = Token(self.__next(TT.DIV).position, TT.STRING, '/')
                    values.append(ast.StringNode(token))

                case TT.MINUS:
                    token = Token(
                        self.__next(
                            TT.MINUS,
                        ).position, TT.STRING, '-',
                    )
                    values.append(ast.StringNode(token))
            nextType = self.__get_next_type()

        return ast.StringExprNode(values)
